#include "chirp_reli.h"
#include "auth_all.h"
#include "debug.h"
#include "chirp_array.h"
#include "xmalloc.h"
#include "nvpair.h"
#include "catalog_query.h"

#include <time.h>
#include <stdio.h>
#include <time.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/wait.h>


#define MAX_SIZE 1048576
#define MAX_FILE 100

int compare_entries( struct nvpair **a, struct nvpair **b )
{
	int result;
	const char *x, *y;

	x = nvpair_lookup_string(*a,"type");
	if(!x) x = "unknown";

	y = nvpair_lookup_string(*b,"type");
	if(!y) y = "unknown";

	result = strcmp(x,y);
	if(result!=0) return result;

        x = nvpair_lookup_string(*a,"name");
        if(!x) x = "unknown";

        y = nvpair_lookup_string(*b,"name");
        if(!y) y = "unknown";

        return strcmp(x,y);
}

struct data_file{
	char host[CHIRP_LINE_MAX];
	char path[CHIRP_LINE_MAX];
};

struct chirp_array {
	char host[CHIRP_LINE_MAX];
	char path[CHIRP_LINE_MAX];
	int width;
	int height;
	int element_size;
	int n_data_files;
	int n_row_per_file;
	struct data_file files[MAX_FILE];
	struct chirp_file *rfiles[MAX_FILE];
	void *cache;
	int rownumber;	
};

/* create a new array at the given path, return zero on failure*/
struct chirp_array * chirp_array_create( const char * host, const char *path, int width, int height, int element_size )
{
	int result;
	static struct nvpair *table[10000];
	char s_flie[CHIRP_LINE_MAX];
	sprintf(s_flie,"%s/.chirp/hosts",getenv("HOME"));
	FILE *in = fopen(s_flie,"r");

	int i;
	struct chirp_array *array;
	time_t stoptime = time(0)+300;



	array = xxmalloc(sizeof(*array));

	strcpy(array->host,host);
	if(!chirp_client_connect(array->host,1,stoptime))
	{
		printf("Error: cannot connect to host %s\n",array->host);
		return 0;
	}
	
	strcpy(array->path,path);
	array->width = width;
	array->height = height;
	array->element_size = element_size;
	array->n_row_per_file = ((double)MAX_SIZE/element_size)*((double)1024/width) + 1;
	array->n_data_files = (double)height/array->n_row_per_file+ 1;
	array->cache = xxmalloc(element_size*width);
	array->rownumber = -1;
	
	FILE *out = fopen("/tmp/c_array.crp","w");
	if(out!=NULL)
		fprintf(out,"%s\n%s\n%d\n%d\n%d\n",host,path,width,height,element_size);
	else
	{
		printf("Error: can not write to /tmp\n");
		return 0;
	}
	
	char PATH[CHIRP_LINE_MAX];
	sprintf(PATH, "%s%s%s%s",path, "/",path,".mtdata");

	struct nvpair *n;
	struct catalog_query *q;
	const char * catalog_host = 0;

	q = catalog_query_create(catalog_host,0,stoptime);
	int count=0;
	while((n = catalog_query_read(q,stoptime)))
	{
		table[count++] = n;
	}
	qsort(table,count,sizeof(*table),(void*)compare_entries);

	for(i = 0;i<array->n_data_files;i++)
	{
			char tempPath[CHIRP_LINE_MAX];
			char tempHost[CHIRP_LINE_MAX];
			char temp[CHIRP_LINE_MAX];
			sprintf(tempPath, "%s%s%s%s%d",path,"/",path,".data", i);
			//if ~/.chirp/.chirp_array exists
			if(in!=NULL)
			{
				do
				{
					if(fscanf(in,"%s",temp)!=EOF)
					{
						sprintf(tempHost, "%s",temp);
						//chirp_reli_mkdir(tempHost,path,0755,stoptime);
						
					}
					else
					{
						rewind(in);
					}
					//printf("%d %s\n",i,tempHost);
				}while(!chirp_client_connect(tempHost,1,stoptime));
			}
			//if not, randomly pick a chirp server from the catalog
			else
			{
				do
				{
					const char *t;
					int i;
					do
					{
						i = rand()%count;
						nvpair_lookup_string(table[i],"type");
						sprintf(tempHost, "%s",nvpair_lookup_string(table[rand()%count],"name"));
					}while(t && !strcmp(t,"chirp"));
					
				}while(!chirp_client_connect(tempHost,1,stoptime));
			}
			result = chirp_reli_mkdir(tempHost,path,0755,stoptime);
			//if(result >= 0)
			//{
				strcpy(array->files[i].host,tempHost);
				strcpy(array->files[i].path,tempPath);
				fprintf(out,"%s %s\n",tempHost,tempPath);
			/*}
			else
			{
				i--;
			}*/
	}

	fclose(out);
	out = fopen("/tmp/c_array.crp","r");
	
	result = chirp_reli_mkdir(host,path,0755,stoptime);

	struct stat64 info;
	fstat64(fileno(out),&info);
	if(chirp_reli_putfile(host,PATH,out,info.st_mode,info.st_size,stoptime)!=info.st_size)
	{
		fprintf(stderr,"couldn't put %s: %s\n",path,strerror(errno));
		return 0;
	}
	for(i = 0;i<array->n_data_files;i++)
	{
			array->rfiles[i] = xxmalloc(sizeof((struct chirp_file*)array->rfiles[i]));
			array->rfiles[i] = chirp_reli_open(array->files[i].host,array->files[i].path,O_RDWR|O_CREAT,0755,stoptime);
	}
	return array;
	
}

/* open an existing array, return zero on failure */
struct chirp_array * chirp_array_open( const char * host, const char *path )
{
	int i;
	struct chirp_array *array;
	array = xxmalloc(sizeof(*array));
	
	time_t stoptime = time(0)+300;
	if(!chirp_client_connect(host,1,stoptime))
	{
		printf("Error: cannot connect to host %s\n",host);
		return 0;
	}

	FILE *tempfile;

	tempfile = fopen("/tmp/tempfile","w");
	if(tempfile==NULL)
	{
		printf("Cannot write to /tmp\n");
		return 0;
	}
	else
	{
		char PATH[CHIRP_LINE_MAX];
		sprintf(PATH, "%s%s%s%s",path, "/",path,".mtdata");
		int result = chirp_reli_getfile(host,PATH,tempfile,stoptime);
		//cannot get metadata file
		if(result<0)
		{
			printf("%s %s %d %i %s chirp_getfile failed\n",host,host,result,errno,strerror(errno));
			return 0;
		}
		else
		{
			fclose(tempfile);
			tempfile = fopen("/tmp/tempfile","r");
			char tmp[CHIRP_LINE_MAX];
			fscanf(tempfile,"%s",tmp);
			strcpy(array->host,tmp);
			fscanf(tempfile,"%s",tmp);
			strcpy(array->path,tmp);
			fscanf(tempfile,"%d",&array->width);
			fscanf(tempfile,"%d",&array->height);
			fscanf(tempfile,"%d",&array->element_size);
			array->n_row_per_file = ((double)MAX_SIZE/array->element_size)*((double)1024/array->width) + 1;
			array->n_data_files = (double)array->height/array->n_row_per_file+ 1;
			array->cache = xxmalloc(array->element_size*array->width);
			array->rownumber = -1;

			for(i = 0;i<array->n_data_files;i++)
			{
				fscanf(tempfile,"%s",tmp);
				strcpy(array->files[i].host,tmp);
				//chirp_reli_mkdir(tmp,array->path,0755,stoptime);
				fscanf(tempfile,"%s",tmp);
				strcpy(array->files[i].path,tmp);
			}
			for(i = 0;i<array->n_data_files;i++)
			{
				array->rfiles[i] = xxmalloc(sizeof((struct chirp_file*)array->rfiles[i]));
				array->rfiles[i] = chirp_reli_open(array->files[i].host,array->files[i].path,O_RDWR|O_CREAT,0755,stoptime);
			}
			return array;
			
		}
	}
}


/* return the proprerties of this currently open array */
int chirp_array_width( struct chirp_array *a )
{
	return a->width;
}

int chirp_array_height( struct chirp_array *a )
{
	return a->height;
}

int chirp_array_element_size( struct chirp_array *a )
{
	return a->element_size;
}
int  chirp_array_get(struct chirp_array *a, int i, int j, void *data )
{
	int result;
	if(i>=a->height)
	{
		printf("row number is out of bound\n");
		return -1;
	}
	if(j>=a->width)
	{
		printf("col number is out of bound\n");
		return -1;
	}

	if(a->rownumber == i)
	{
		memcpy(data,a->cache+j*a->element_size,a->element_size);
		return 0;
	}
	else
	{
		time_t stoptime = time(0)+300;
		int index = i/a->n_row_per_file;
		INT64_T offset= ((i%a->n_row_per_file)*a->width + j)*a->element_size;
		chirp_reli_pread_unbuffered(a->rfiles[index],data,sizeof(data),offset,stoptime);
		a->rownumber = i;
		return 0;
	}
}


/* get the indicated element and copy it to the pointer data */
int  chirp_array_get_buffered( struct chirp_array *a, int i, int j, void *data )
{
	int result;
	if(i>=a->height)
	{
		printf("row number is out of bound\n");
		return -1;
	}
	if(j>=a->width)
	{
		printf("col number is out of bound\n");
		return -1;
	}

	if(a->rownumber == i)
	{
		memcpy(data,a->cache+j*a->element_size,a->element_size);
		return 0;
	}
	else
	{
		time_t stoptime = time(0)+300;
		int index = i/a->n_row_per_file;

		INT64_T offset= (i%a->n_row_per_file)*a->width*a->element_size;
		result = chirp_reli_pread(a->rfiles[index],a->cache,a->element_size*a->width,offset,stoptime);
		if(result<0)
		{
			printf("error: %s\n",strerror(errno));
			return -1;
		}
		memcpy(data,a->cache+j*a->element_size,a->element_size);
		a->rownumber = i;
		return 0;
	}
}

/*
int  chirp_array_get( struct chirp_array *a, int i, int j, void *data )
{
	time_t stoptime = time(0)+300;
	int index = i/a->n_row_per_file;
	INT64_T offset= ((i%a->n_row_per_file)*a->width + j)*a->element_size;
	chirp_reli_pread(a->rfiles[index],data,sizeof(data),offset,stoptime);
	a->rownumber = i;
	return 0;
}*/

/* get the indicated row and copy it to the pointer data */
int  chirp_array_get_row( struct chirp_array *a, int i, void *data )
{
	if(i>=a->height)
	{
		printf("row number is out of bound\n");
		return -1;
	}
	int result;
	time_t stoptime = time(0)+300;
	int index = i/a->n_row_per_file;

	INT64_T offset= (i%a->n_row_per_file)*a->width*a->element_size;
	result = chirp_reli_pread(a->rfiles[index],data,a->element_size*a->width,offset,stoptime);
	if(result<0)
	{
		printf("error: %s\n",strerror(errno));
		return -1;
	}
	return 0;
}

/* set the indicated element by copying it to the pointer data */
int  chirp_array_set( struct chirp_array *a, int i, int j, const void *data )
{
	if(i>=a->height)
	{
		printf("row number is out of bound\n");
		return -1;
	}
	if(j>=a->width)
	{
		printf("col number is out of bound\n");
		return -1;
	}

	int result;
	time_t stoptime = time(0)+300;
	int index = i/a->n_row_per_file;

	INT64_T offset= ((i%a->n_row_per_file)*a->width + j)*a->element_size;
	result = chirp_reli_pwrite(a->rfiles[index],data,sizeof(data),offset,stoptime);
	if(result<0)
	{
		printf("error: %s\n",strerror(errno));
		return -1;
	}

	//printf("%lf what\n",((double *)(data))[0]);
	//invalidate cache
	if(a->rownumber == i)
		a->rownumber = -1;
	return 0;
	
}

/* set the indicated row by copying it to the pointer data */
int  chirp_array_set_row( struct chirp_array *a, int i, const void *data )
{
	if(i>=a->height)
	{
		printf("row number is out of bound\n");
		return -1;
	}

	int result;
	time_t stoptime = time(0)+300;
	int index = i/a->n_row_per_file;
	INT64_T offset= (i%a->n_row_per_file)*a->width*a->element_size;
	result = chirp_reli_pwrite(a->rfiles[index],data,a->element_size*a->width,offset,stoptime);
	if(result<0)
	{
		printf("error: %s\n",strerror(errno));
		return -1;
	}
	//invalidate cache
	if(a->rownumber == i)
		a->rownumber = -1;
	return 0;
}

int  chirp_array_get_col( struct chirp_array *a, int i,void *data )
{
	if(i>=a->width)
	{
		printf("col number is out of bound\n");
		return -1;
	}
	int result;
	int j;
	time_t stoptime = time(0)+300;
	
	INT64_T offset= i*a->element_size;
	int length = a->element_size*a->n_row_per_file;
	
	for(j = 0;j<a->n_data_files-1;j++)
	{
		result = chirp_reli_sread(a->rfiles[j], data+j*length, (a->height - a->n_row_per_file*j)*a->element_size, a->element_size, a->element_size*a->width, offset,stoptime);
		if(result<0)
		{
			printf("error: %s\n",strerror(errno));
			return -1;
		}
	}
	result = chirp_reli_sread(a->rfiles[j], data+j*length, (a->height - a->n_row_per_file*j)*a->element_size, a->element_size, a->element_size*a->width, offset,stoptime);
	if(result<0)
	{
		printf("error: %s\n",strerror(errno));
		return -1;
	}
	return 0;

}

int chirp_array_set_col( struct chirp_array *a, int i,const void *data )
{
	if(i>=a->width)
	{
		printf("col number is out of bound\n");
		return -1;
	}
	int result;
	int j;
	time_t stoptime = time(0)+300;
	
	INT64_T offset= i*a->element_size;
	int length = a->element_size*a->n_row_per_file;
	
	for(j = 0;j<a->n_data_files-1;j++)
	{
		result = chirp_reli_swrite(a->rfiles[j], data+j*length, (a->height - a->n_row_per_file*j)*a->element_size, a->element_size, a->element_size*a->width, offset,stoptime);
		if(result<0)
		{
			printf("error: %s\n",strerror(errno));
			return -1;
		}
	}
	result = chirp_reli_swrite(a->rfiles[j], data+j*length, (a->height - a->n_row_per_file*j)*a->element_size, a->element_size, a->element_size*a->width, offset,stoptime);
	if(result<0)
	{
		printf("error: %s\n",strerror(errno));
		return -1;
	}
	//invalidate cache
	a->rownumber = -1;
	return 0;
}

/* close this array and free all related resources */
void chirp_array_close( struct chirp_array *a )
{
	free(a);
}
